#include "types.h"

class CPagesList {
  struct CPagesItem {
   DWord Line;
   CPagesItem *Next;
  };//CPagesItem
  CPagesItem *Head;
 public:
  CPagesList ();
  ~CPagesList ();
  int Empty ();
  DWord Count ();
  int Search (DWord line);
  void InsertEnd (DWord line);
  DWord GetItem (DWord index);
};//CPagesList

class CWordTree {
  char *w;
  char wlen;
  struct CWordNode {
   char Letter;
   CPagesList PagesList;
   CWordNode *Sons[MaxChars];
  };//CWordNode
  CWordNode *Root;
  void DestroyTree (CWordNode *&p);
  void GoThroughRec (CWordNode *p, Procedure proc);
 public:
  CWordTree ();
  ~CWordTree ();
  void Insert (char *_word, char wordlen, DWord lineindex);
  void GoThrough (Procedure proc);
};//CWordTree

CPagesList::CPagesList ()
{
 Head = NULL;
}//CPagesList

CPagesList::~CPagesList ()
{
 CPagesItem *p;
 while (Head != NULL) {
  p = Head;
  Head = Head->Next;
  delete p;
 }//while
}//~CPagesList

int CPagesList::Empty ()
{
 if (Head == NULL) return 1; else return 0;
}//Empty

DWord CPagesList::Count ()
{
 DWord countnum = 0;
 CPagesItem *p = Head;
 while (p != NULL) {
  countnum++;
  p = p->Next;
 }//while
 return countnum;
}//Count

int CPagesList::Search (DWord line)
{
 CPagesItem *p = Head;
 while (p != NULL) {
  if (p->Line == line)
   return 1;
  p = p->Next;
 }//while
 return 0;
}//Search

void CPagesList::InsertEnd (DWord line)
{
 CPagesItem *p = Head, *p0 = p;
 while (p != NULL) {
  p0 = p;
  p = p->Next;
 }//while
 if (!Search (line)) {
  p = new CPagesItem;
  if (p == NULL) ErrorMsg (ErrorMessages[0]);
  p->Line = line;
  p->Next = NULL;
  if (Head == NULL) Head = p; else p0->Next = p;
 }//if
}//InsertEnd

DWord CPagesList::GetItem (DWord index)
{
 CPagesItem *p = Head;
 for (int i = 0; i < index; i++)
  p = p->Next;
 return p->Line;
}//Get

void CWordTree::DestroyTree (CWordNode *&p)
{
 if (p != NULL) {
  for (int i = 0; i < MaxChars; i++)
   DestroyTree (p->Sons[i]);
  delete p;
 }//if
}//DestroyTree

void CWordTree::GoThroughRec (CWordNode *p, Procedure proc)
{
 if (p != NULL) {
  w[wlen] = p->Letter;
  wlen++;
  if (!(p->PagesList.Empty ())) {
   DWord num = p->PagesList.Count ();
   DWord *listarr = (DWord *)malloc ((Word)num);
   if (listarr == NULL) ErrorMsg (ErrorMessages[0]);
   for (int j = 0; j < num; j++)
    listarr[j] = p->PagesList.GetItem (j);
   proc (w, wlen, listarr, num);
   free (listarr);
  }//if
  for (int i = 0; i < MaxChars; i++)
   GoThroughRec (p->Sons[i], proc);
  wlen--;
 }//if
}//GoThroughRec

CWordTree::CWordTree ()
{
 Root = new CWordNode;
 if (Root == NULL) ErrorMsg (ErrorMessages[0]);
 for (int i = 0; i < MaxChars; i++)
  Root->Sons[i] = NULL;
}//CWordTree

CWordTree::~CWordTree ()
{
 DestroyTree (Root);
}//~CWordTree

void CWordTree::Insert (char *_word, char wordlen, DWord lineindex)
{
 CWordNode *p = Root, *p0;
 for (int i = 0; i < wordlen; i++) {
  p0 = p;
  if (p->Sons[_word[i] - A] == NULL) {
   p = new CWordNode;
   if (p == NULL) ErrorMsg (ErrorMessages[0]);
   p->Letter = _word[i];
   for (int j = 0; j < MaxChars; j++)
    p->Sons[j] = NULL;
   p0->Sons[_word[i] - A] = p;
  }//if
  else p = p->Sons[_word[i] - A];
 }//for
 p->PagesList.InsertEnd (lineindex);
}//Insert

void CWordTree::GoThrough (Procedure proc)
{
 for (int i = 0; i < MaxChars; i++) {
  w = "";
  wlen = 0;
  GoThroughRec (Root->Sons[i], proc);
 }//for
}//GoThrough